/*
Copyright (C)2014 Raster Software Vigo (Sergio Costas)

This file is part of launch_debian

Launch_debian is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

Launch_debian is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <unistd.h>
#include <sys/inotify.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/time.h>


// Mount point for the chrooted system
#define LAUNCH_DEBIAN_MOUNT_PATH "/system/debian"
// Where to search for new devices
#define LAUNCH_DEBIAN_CHECK_PATH "/dev/block"
// Type of files to check as valid devices
#define LAUNCH_DEBIAN_DEV_LIST {"sd","mmcblk",NULL}
// Default mount type
#define LAUNCH_DEBIAN_MOUNT_TYPE "ext4"
// Default FIFO path
#define LAUNCH_DEBIAN_CONTROL_FIFO "/dev/chroot_ext_control"


char *mount_path = LAUNCH_DEBIAN_MOUNT_PATH;
char *check_path = LAUNCH_DEBIAN_CHECK_PATH;
char *mount_options = NULL;
char *mount_type = LAUNCH_DEBIAN_MOUNT_TYPE;
char *control_fifo = LAUNCH_DEBIAN_CONTROL_FIFO;

int inotify_dsc;
int path_dsc;

#define LAUNCH_DEBIAN_BUFFER_SIZE 1024

int init_inotify() {

	inotify_dsc = inotify_init();
	if (inotify_dsc < 0) {
		return -1;
	}
	
	path_dsc = inotify_add_watch(inotify_dsc,check_path,IN_CREATE);
	if (path_dsc < 0) {
		return -1;
	}
	return 0;
}

void killall() {

	char command[1024];

	while(1) {
		sprintf(command,"busybox fuser -k %s",mount_path);
		system(command);
		sprintf(command,"mount -o remount,ro %s",mount_path);
		if (0 == system(command)) {
			break;
		}
		sleep(1);
	}
}

void run_command(char *run_command) {

	char command[1024];

	if (!strcmp(run_command,"reboot")) {
		killall();
		system(command);
		system("reboot");
		return;
	}
	if (!strcmp(run_command,"halt")) {
		killall();
		return;
	}
}

void check_commands() {
	
	int fifofd=-1;
	int retval;
	char buffer[1025];
	char *p,*command;
	
	mkfifo(control_fifo,S_IRWXU|S_IRWXG);
	
	while(1) {
		if (fifofd==-1) {
			fifofd = open(control_fifo,O_RDONLY);
			if (fifofd < 0) {
				printf("Can't open control FIFO. Aborting\n");
				exit(1);
			}
		}
		retval = read(fifofd,buffer,1024);
		if (retval<=0) {
			close(fifofd);
			fifofd=-1;
			continue;
		}
		buffer[retval]=0;
		for(p=buffer,command=buffer;*p!=0;p++) {
			if (*p<32) {
				if (*p != 0) {
					*p=0;
					p++;
				}
				if (*command != 0) {
					run_command(command);
				}
				command=p;
				continue;
			}
		}
	}
}

// Checks if the device pointed by NAME is a valid EXT4 partition and contains a valid runlinux.sh file
void check_device(char *name) {

	char **pointer;
	char *elements[] = LAUNCH_DEBIAN_DEV_LIST;
	char command[1024];

	for (pointer=elements;*pointer!=NULL;pointer++) {
		if (0==strncmp(name,*pointer,strlen(*pointer))) {
			// it it matches, we found a valid device
			sprintf(command,"mount -t %s %s %s %s/%s %s",
				mount_type,
				mount_options == NULL ? "" : "-o",
				mount_options == NULL ? "" : mount_options,
				check_path,name,mount_path);
			// try to mount it as an ext4 filesystem
			if (0 != system(command)) {
				break; // if it fails, it is not an ext4 file system
			}
			sprintf(command,"%s/runlinux.sh %s %s",mount_path,mount_path,name);
			if(system(command) == 0) {
				check_commands(); // launched the desired command, so we start checking the PIPE for chroot commands
			}
			// after running the file (if it doesn't exists, or it is no valid), unmount the file system
			sprintf(command,"umount %s",mount_path);
			system(command);
			break;
		}
	}
}

void show_options() {

	printf("\nUsage: launch_debian [-o mount_options] [-m mount_path] [-t filesystem_type] [-c check_path_for_devices] [-f path_for_control_fifo]\n");
	printf("Default mount path: %s\n",LAUNCH_DEBIAN_MOUNT_PATH);
	printf("Default filesystem type: %s\n",LAUNCH_DEBIAN_MOUNT_TYPE);
	printf("Default path for devices: %s\n",LAUNCH_DEBIAN_CHECK_PATH);
	printf("Default path for control FIFO: %s\n",LAUNCH_DEBIAN_CONTROL_FIFO);
	printf("\n");
}

int check_arg(int counter,char **argv) {

	char *param = argv[counter];
	if (!strcmp(param,"-o")) {
		mount_options = argv[counter+1];
		return (counter+2);
	}
	if (!strcmp(param,"-m")) {
		mount_path = argv[counter+1];
		return (counter+2);
	}
	if (!strcmp(param,"-t")) {
		mount_type = argv[counter+1];
		return (counter+2);
	}
	if (!strcmp(param,"-c")) {
		check_path = argv[counter+1];
		return (counter+2);
	}
	if (!strcmp(param,"-f")) {
		control_fifo = argv[counter+1];
		return (counter+2);
	}
	printf("Error: parameter %s unknown.\n",param);
	show_options();
	exit(1);
}

int main(int argc, char **argv) {

	char buffer[LAUNCH_DEBIAN_BUFFER_SIZE];
	struct inotify_event *event;
	int retval;
	int position;

	if (argc>1) {
		int counter=1;
		while(counter<=argc) {
			counter += check_arg(counter,argv);
			if (counter>(argc+1)) {
				printf("Incomplete parameter\n");
				show_options();
				exit(1);
			}
		}
	}

	if(init_inotify() !=0) {
		return -1;
	}

	// check all the currently existing devices
	// to allow it to work from start
	DIR *directory = opendir(check_path);
	if (directory != NULL) {
		struct dirent *entry;
		do {
			entry = readdir(directory);
			if (entry != NULL) {
				check_device(entry->d_name);
			}
		} while (entry != NULL);
		closedir(directory);
	}

	while(1) {
		retval = read(inotify_dsc,buffer, LAUNCH_DEBIAN_BUFFER_SIZE);
		if (retval <= 0) {
			continue;
		}
		for (position=0; position < retval;) {
			event = (struct inotify_event *)(buffer + position);
			position += event->len + sizeof(struct inotify_event);
			check_device(event->name);
		}
	}
}
